#include <iostream>

#include "glpanel.h"

#ifndef ABS
#define ABS(x) (((x) < 0.0)? -(x): (x))
#endif

//--------------------------------------------------------------------
GLPanel::GLPanel(QWidget *parent)
 : QGLWidget(parent)
{
 setMouseTracking(true);
 setFocusPolicy(Qt::ClickFocus);
 m = NULL;
 model_dsp_id = 0;
 updated_model_dsp = false;
 left = -1.0;
 right = 11.0;
 bottom = -1.0;
 top = 11.0;
 curMouseAction = GLPanel::SELECTION;
 button_pressed = false;
 mouse_button = Qt::NoButton;
 pt0.setX(0);
 pt0.setY(0);
 pt1.setY(0);
 pt1.setY(0);
 pick_tol_fac = 0.01;
 mouse_move_tol = 2;
 shift_key_pressed = false;
 colorBackground   = Qt::white;
 colorCollecting   = Qt::darkMagenta;
 colorCurve        = Qt::darkYellow;
 colorVertex       = Qt::darkCyan;
 colorSelection    = Qt::red;
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
GLPanel::~GLPanel()
{
 if(!m)
  delete m;
 if(model_dsp_id > 0)
  glDeleteLists(model_dsp_id, 1);
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::setModel(GeomModel* model)
{
 if(!m)
  delete m;
 m = model;
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::setMouseAction( int _action )
{
 if( curMouseAction == _action )
  return;

 switch( _action )
 {
  default:
  case GLPanel::SELECTION:
   curMouseAction = GLPanel::SELECTION;
   collector.reset();
   break;
  case GLPanel::COLLECTION:
   curMouseAction = GLPanel::COLLECTION;
   break;
 }
 paintGL();
 update();
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::setCurveType( int _type )
{
 if( (curMouseAction == GLPanel::COLLECTION) &&
     (collector.getCurveType() == _type) )
  return;

 collector.reset();
 collector.setCurveType( _type );
 paintGL();
 update();
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::delSelectedCurves()
{
 if( (m != NULL) && (!m->isEmpty()) )
 {
  m->delSelectedCurves();
  updated_model_dsp = false;
  paintGL();
  update();
 }
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::initializeGL()
{
 // set white as background color and clear window
 qglClearColor(colorBackground);
 glClear(GL_COLOR_BUFFER_BIT);
 glEnable(GL_LINE_SMOOTH);
}

//--------------------------------------------------------------------
void 
GLPanel::resizeGL(int width, int height)
{
 // Avoid division by zero
 if(height == 0) height = 1;

 // store GL canvas sizes in object properties
 w = width;
 h = height;

 // setup world space window limits based on object bounding box
 if( (m != NULL) && (!m->isEmpty()) )
 {
  m->getBoundBox(&left, &right, &bottom, &top);
  scaleWorldWindow(1.10);
 }
 else
 {
  scaleWorldWindow(1.0);
 }

 // setup the viewport to canvas dimensions
 glViewport(0, 0, (GLint)w, (GLint)h);

 // reset the coordinate system
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();

 // establish the clipping volume by setting up an orthographic projetction
 glOrtho(left, right, bottom, top, -1.0, 1.0);

 // setup display in model coordinates
 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity();
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::paintGL()
{
 // clear the buffer with the current clear color
 glClear(GL_COLOR_BUFFER_BIT);

 // display model (if there is one)
 if( (m != NULL) && (!m->isEmpty()) )
 {
  if(!updated_model_dsp)
  {
   if(model_dsp_id > 0)
    glDeleteLists(model_dsp_id, 1);
   model_dsp_id = makeDisplayModel();
  }
  if(model_dsp_id > 0)
  {
   glCallList(model_dsp_id);
   updated_model_dsp = true;
  }
 }

 switch( curMouseAction )
 {
  case GLPanel::SELECTION:
   {
    // Check to see whether there is selection fence and,
    // if that is the case, draw it
    GLuint col_fence = drawSelectionFence();
    if(col_fence > 0)
    {
     glCallList(col_fence);
     glDeleteLists(col_fence, 1);
    }
   }
   break;
  case GLPanel::COLLECTION:
   {
    // Check to see whether there is a curve being collected and,
    // if that is the case, draw it
    GLuint col_curve = drawCollectedCurve();
    if(col_curve > 0)
    {
     glCallList(col_curve);
     glDeleteLists(col_curve, 1);
    }
   }
  break;
 }
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
GLuint 
GLPanel::makeDisplayModel()
{
 if( (m == NULL) || m->isEmpty() )
  return 0;

 GLuint list = glGenLists(1);
 glNewList(list, GL_COMPILE);

  // view the model
 glLineWidth(1.0);
 vector<Curve*> pCurves = m->getCurves();
 for( int i = 0; i < pCurves.size(); i++ )
 {
  vector<Point> pts = pCurves[i]->getPointsToDraw();

  // display curve lines
  if( pCurves[i]->isSelected() )
   qglColor( colorSelection );
  else
   qglColor( colorCurve );
  glBegin(GL_LINE_STRIP);
  for( int j = 0; j < pts.size(); j++ )
  {
   glVertex2d((GLdouble)pts[j].getX(),
              (GLdouble)pts[j].getY());
  }
  glEnd();

  // display curve vertices
  glPointSize(5.0);
  qglColor( colorVertex );
  glBegin(GL_POINTS);
  glVertex2d((GLdouble)pts[0].getX(),
             (GLdouble)pts[0].getY());
  glVertex2d((GLdouble)pts[pts.size()-1].getX(),
             (GLdouble)pts[pts.size()-1].getY());
  glEnd();
 }
 glEndList();
 return list;
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
GLuint 
GLPanel::drawSelectionFence()
{
 // It is assumed that the current mouse action is for curve selection

 // Only draw a fence if mouse button is pressed
 if( !button_pressed )
  return 0;

 // If current mouse point is in the same position of inital mouse point
 // do not display anything
 if( (pt0.x() == pt1.x()) && (pt0.y() == pt1.y()) )
  return 0;

// If there is no model or it is empty, do nothing
 if( (m == NULL) || m->isEmpty() )
  return 0;

 // Display a fence from initial mouse point to current mouse point
 GLuint list = glGenLists(1);
 glNewList(list, GL_COMPILE);

 glLineWidth(1.0);
 qglColor( colorSelection );

 glBegin(GL_LINE_STRIP);
 QPointF pt0W = convertPtCoordsToUniverse( pt0 );
 QPointF pt1W = convertPtCoordsToUniverse( pt1 );
 glVertex2d((GLdouble)pt0W.x(), (GLdouble)pt0W.y());
 glVertex2d((GLdouble)pt1W.x(), (GLdouble)pt0W.y());
 glVertex2d((GLdouble)pt1W.x(), (GLdouble)pt1W.y());
 glVertex2d((GLdouble)pt0W.x(), (GLdouble)pt1W.y());
 glVertex2d((GLdouble)pt0W.x(), (GLdouble)pt0W.y());
 glEnd();

 glEndList();
 return list;
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
GLuint 
GLPanel::drawCollectedCurve()
{
 // It is assumed that the current mouse action if for curve collection

 if( (!collector.isActive()) || (!collector.isCollecting()) )
  return 0;

 GLuint list = glGenLists(1);
 glNewList(list, GL_COMPILE);

 glLineWidth(1.0);
 qglColor( colorCollecting );

 // Display lines of curve being collected
 glBegin(GL_LINE_STRIP);
 vector<Point> pts = collector.getDrawPoints();
 for( int j = 0; j < pts.size(); j++ )
 {
  double x = pts[j].getX();
  double y = pts[j].getY();
  glVertex2d((GLdouble)x, (GLdouble)y);
 }
 glEnd();

 // Display control points of curve being collected
 glPointSize(5.0);
 glBegin(GL_POINTS);
 vector<Point> ctrl_pts = collector.getPoints();
 for( int j = 0; j < ctrl_pts.size(); j++ )
 {
  double x = ctrl_pts[j].getX();
  double y = ctrl_pts[j].getY();
  glVertex2d((GLdouble)x, (GLdouble)y);
 }
 glEnd();

 glEndList();
 return list;
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::scaleWorldWindow(double scale_fac)
{
 double vpr;          // viewport distortion ratio
 double cx, cy;       // window center
 double sizex, sizey; // window sizes

 //Compute canvas viewport ratio.
 vpr = (double)h / (double)w;

 // Get current window center.
/*** COMPLETE AQUI: 01 ***/


/*** COMPLETE AQUI: 01 ***/

// Set new window sizes based on scaling factor.
/*** COMPLETE AQUI: 02 ***/


/*** COMPLETE AQUI: 02 ***/

// Adjust window to keep the same aspect ratio of the viewport.
/*** COMPLETE AQUI: 03 ***/









/*** COMPLETE AQUI: 03 ***/

 // Establish the clipping volume by setting up an orthographic projetction
 glMatrixMode(GL_PROJECTION); 
 glLoadIdentity();
 glOrtho(left, right, bottom, top, -1.0, 1.0);
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::panWorldWindow(double pan_facX, double pan_facY)
{
 // Compute pan distances in horizontal and vertical directions.
 double panX = (right - left)   * pan_facX;
 double panY = (top   - bottom) * pan_facY;

 // Shift current window.
/*** COMPLETE AQUI: 04 ***/




/*** COMPLETE AQUI: 04 ***/

 // Establish the clipping volume by setting up an orthographic projetction
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 glOrtho(left, right, bottom, top, -1.0, 1.0);
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::fitWorldToViewport()
{
 if(!m)
  return;
 m->getBoundBox(&left, &right, &bottom, &top);
 scaleWorldWindow(1.10);
 paintGL();
 update();
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::zoomIn()
{
 if(!m)
  return;
/*** COMPLETE AQUI: 05 ***/



/*** COMPLETE AQUI: 05 ***/
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::zoomOut()
{
 if(!m)
  return;
/*** COMPLETE AQUI: 06 ***/



/*** COMPLETE AQUI: 06 ***/
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::panLeft()
{
 if(!m)
  return;
/*** COMPLETE AQUI: 07 ***/



/*** COMPLETE AQUI: 07 ***/
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::panRight()
{
 if(!m)
  return;
/*** COMPLETE AQUI: 08 ***/



/*** COMPLETE AQUI: 08 ***/
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::panDown()
{
 if(!m)
  return;
/*** COMPLETE AQUI: 09 ***/



/*** COMPLETE AQUI: 09 ***/
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::panUp()
{
 if(!m)
  return;
/*** COMPLETE AQUI: 10 ***/



/*** COMPLETE AQUI: 10 ***/
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
QPointF 
GLPanel::convertPtCoordsToUniverse(QPoint _pt)
{
 double dX = right - left; // universe window horizontal size
 double dY = top - bottom; // universe window vertical size

 // Origin of canvas raster coordinates it at the left-top corner,
 // while origin of GL canvas floating point coordinates is at
 // left-bottom corner.
 // mX is the distance of point to left universe window limit
 // mY is the distance of point to bottom universe window limit
 double mX = (double)_pt.x()*dX/(double)w;
 double mY = (double)(h - _pt.y())*dY/(double)h; 

 double x = left + mX;
 double y = bottom + mY;
 return QPointF(x,y);
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
//                      MOUSE EVENT SLOTS
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::mousePressEvent(QMouseEvent* event)
{
 //////////////////////////////////////////
 // std::cout << ".MOUSE_PRESS " << std::endl;
 //////////////////////////////////////////
 button_pressed = true;
 pt0 = event->pos();
 mouse_button = event->button();
 switch( curMouseAction )
 {
  default:
  case GLPanel::SELECTION:
   break;
  case GLPanel::COLLECTION:
   if( !collector.isActive() )
   {
    // In case of left mouse button, if not doing any curve collection,
    // start a new curve collection
    if( mouse_button == Qt::LeftButton )
    {
     collector.startCurveCollection();
    }
   }
   break;
 }
}

//--------------------------------------------------------------------
void 
GLPanel::mouseMoveEvent(QMouseEvent* event)
{
 // Get current mouse position
 pt1 = event->pos();

 // Treat mouse move event according to current mouse action type
 switch( curMouseAction )
 {
  default:
   break;
  case GLPanel::SELECTION:
   // Disregard mouse move event if left mouse button is not pressed
   if( (mouse_button == Qt::LeftButton) && button_pressed )
   {
    paintGL();
    update();
   }
   break;
  case GLPanel::COLLECTION:
   // Only consider current point if left mouse button was used,
   // if not button pressed, and if current point is not at the
   // same location of button press point.
   if( mouse_button == Qt::LeftButton && (!button_pressed) )
   {
    if( (ABS(pt0.x() - pt1.x()) > mouse_move_tol) ||
        (ABS(pt0.y() - pt1.y()) > mouse_move_tol) )
    {
     if( collector.isCollecting() )
     {
      // Convert current mouse position to world coordinates
      // and add it as a temporary point for curve collection.
      QPointF pt1W = convertPtCoordsToUniverse( pt1 );
      collector.addTempPoint(pt1W.x(), pt1W.y());
      paintGL();
      update();
     }
    }
   }
   break;
 }
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::mouseReleaseEvent(QMouseEvent *event)
{
 button_pressed = false;

 // Get current mouse position
 pt1 = event->pos();

 // Convert mouse positions to world coordinates
 QPointF pt0W = convertPtCoordsToUniverse( pt0 );
 QPointF pt1W = convertPtCoordsToUniverse( pt1 );

 // Treat mouse release event according to current mouse action type
 switch( curMouseAction )
 {
  default:
   break;
  case GLPanel::SELECTION:
   if( mouse_button == Qt::LeftButton )
   {
    // If there is a model and it is not empty
    if( (m != NULL) && (!m->isEmpty()) )
    {
     // If current mouse point is in the same position of inital mouse point,
     // consider a mouse pick selection.
     // Otherwise, consider a mouse fence selection.
     if( (ABS(pt0.x() - pt1.x()) <= mouse_move_tol) &&
         (ABS(pt0.y() - pt1.y()) <= mouse_move_tol) )
     {
      double max_size = ((right-left) > (top-bottom) ) ?
                         (right-left) : (top-bottom);
      double pick_tol = max_size * pick_tol_fac;
      m->selectPick( pt1W.x(), pt1W.y(), pick_tol, shift_key_pressed );
     }
     else
     {
      double xmin = (pt0W.x() < pt1W.x()) ? pt0W.x() : pt1W.x();
      double xmax = (pt0W.x() > pt1W.x()) ? pt0W.x() : pt1W.x();
      double ymin = (pt0W.y() < pt1W.y()) ? pt0W.y() : pt1W.y();
      double ymax = (pt0W.y() > pt1W.y()) ? pt0W.y() : pt1W.y();
      m->selectFence( xmin, xmax, ymin, ymax, shift_key_pressed );
     }
    }
    updated_model_dsp = false;
    paintGL();
    update();
   }
   break;
  case GLPanel::COLLECTION:
   // Only consider current point if left mouse point was used and
   // if it is on the same location of button press point.
   if( mouse_button == Qt::LeftButton )
   {
    if( (ABS(pt0.x() - pt1.x()) <= mouse_move_tol) &&
        (ABS(pt0.y() - pt1.y()) <= mouse_move_tol) )
    {
     collector.insertPoint(pt1W.x(), pt1W.y());
    }
   }

   // Check for end of curve collection, which occurs in two situations:
   // (a) If left mouse button was used, and current collected curve has
   //     a limit in the number of points, verify whether current point
   //     finishes collection of current curve.
   // (b) If right mouse button was used, and current collected curve has
   //     no limit in the number of points, verify whether previously
   //     collected points can finish collection of current curve.
   bool endCollection = false;
   if( mouse_button == Qt::LeftButton )
   {
    if( !collector.isUnlimited() )
    {
     if( collector.hasFinished() )
     {
      endCollection = true;
     }
    }
   }
   else if( mouse_button == Qt::RightButton )
   {
    if( collector.isUnlimited() )
    {
     if( collector.hasFinished() )
     {
      endCollection = true;
     }
     else
     {
      // If right mouse button was used, curve has an unlimited number of points, and
      // previously collected points cannot finish collection of curve, reset current
      // curve collection
      collector.reset();
      paintGL();
      update();
     }
    }
    else
    {
     // If right mouse button was used and curve has a limited number of points,
     // reset current curve collection
     collector.reset();
     paintGL();
     update();
    }
   }
   if( endCollection )
   {
    Curve* curve = collector.getCollectedCurve();
    m->insertCurve( curve );
    collector.endCurveCollection();
    updated_model_dsp = false;
    paintGL();
    update();
   }
   break;
 }
}

//--------------------------------------------------------------------
void 
GLPanel::keyPressEvent( QKeyEvent * event )
{
 shift_key_pressed = (event->key() == Qt::Key_Shift);
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GLPanel::keyReleaseEvent( QKeyEvent * event )
{
 shift_key_pressed = false;
}
//--------------------------------------------------------------------
